@@ -5,24 +5,16 @@
   which can be found at http://www.daemonology.net/bsdiff.
 */
 
+#define PY_SSIZE_T_CLEAN
 #include <Python.h>
 
-#if PY_MAJOR_VERSION >= 3
-#define IS_PY3K
+#if PY_MAJOR_VERSION == 2
+#define Py_MIN(x, y)  (((x) > (y)) ? (y) : (x))
 #endif
 
-#ifdef IS_PY3K
-#include "bytesobject.h"
-#define PyString_FromStringAndSize  PyBytes_FromStringAndSize
-#define PyString_Check  PyBytes_Check
-#define PyString_Size  PyBytes_Size
-#define PyString_AsString  PyBytes_AsString
-#endif
 
-#define MIN(x, y)  (((x) < (y)) ? (x) : (y))
-
-
-static void split(off_t *I, off_t *V, off_t start, off_t len, off_t h)
+static void
+split(off_t *I, off_t *V, off_t start, off_t len, off_t h)
 {
     off_t i, j, k, x, tmp, jj, kk;
 
@@ -105,7 +97,8 @@ static void split(off_t *I, off_t *V, off_t start, off_t len, off_t h)
 }
 
 
-static void qsufsort(off_t *I, off_t *V, unsigned char *old, off_t oldsize)
+static void
+qsufsort(off_t *I, off_t *V, unsigned char *old, off_t oldsize)
 {
     off_t buckets[256], i, h, len;
 
@@ -138,7 +131,7 @@ static void qsufsort(off_t *I, off_t *V, unsigned char *old, off_t oldsize)
                 i -= I[i];
             } else {
                 if (len)
-                    I[i - len] =- len;
+                    I[i - len] = -len;
                 len = V[I[i]] + 1 - i;
                 split(I, V, i, len, h);
                 i += len;
@@ -146,7 +139,7 @@ static void qsufsort(off_t *I, off_t *V, unsigned char *old, off_t oldsize)
             }
         }
         if (len)
-            I[i - len] =- len;
+            I[i - len] = -len;
     }
 
     for (i = 0; i < oldsize + 1; i++)
@@ -154,8 +147,9 @@ static void qsufsort(off_t *I, off_t *V, unsigned char *old, off_t oldsize)
 }
 
 
-static off_t matchlen(unsigned char *old, off_t oldsize,
-                      unsigned char *new, off_t newsize)
+static off_t
+matchlen(unsigned char *old, off_t oldsize,
+         unsigned char *new, off_t newsize)
 {
     off_t i;
 
@@ -166,10 +160,11 @@ static off_t matchlen(unsigned char *old, off_t oldsize,
 }
 
 
-static off_t search(off_t *I,
-                    unsigned char *old, off_t oldsize,
-                    unsigned char *new, off_t newsize,
-                    off_t st, off_t en, off_t *pos)
+static off_t
+search(off_t *I,
+       unsigned char *old, off_t oldsize,
+       unsigned char *new, off_t newsize,
+       off_t st, off_t en, off_t *pos)
 {
     off_t x, y;
 
@@ -187,7 +182,7 @@ static off_t search(off_t *I,
     }
 
     x = st + (en - st) / 2;
-    if (memcmp(old + I[x], new, MIN(oldsize - I[x], newsize)) < 0) {
+    if (memcmp(old + I[x], new, Py_MIN(oldsize - I[x], newsize)) < 0) {
         return search(I, old, oldsize, new, newsize, x, en, pos);
     } else {
         return search(I, old, oldsize, new, newsize, st, x, pos);
@@ -198,12 +193,13 @@ static off_t search(off_t *I,
 /* performs a diff between the two data streams and returns a tuple
    containing the control, diff and extra blocks that bsdiff produces
 */
-static PyObject* diff(PyObject* self, PyObject* args)
+static PyObject *
+diff(PyObject* self, PyObject* args)
 {
     off_t lastscan, lastpos, lastoffset, oldscore, scsc, overlap, Ss, lens;
     off_t *I, *V, dblen, eblen, scan, pos, len, s, Sf, lenf, Sb, lenb, i;
     PyObject *controlTuples, *tuple, *results, *temp;
-    int origDataLength, newDataLength;
+    Py_ssize_t origDataLength, newDataLength;
     char *origData, *newData;
     unsigned char *db, *eb;
 
@@ -257,6 +253,7 @@ static PyObject* diff(PyObject* self, PyObject* args)
     lastscan = 0;
     lastpos = 0;
     lastoffset = 0;
+    pos = 0;
     while (scan < newDataLength) {
         oldscore = 0;
 
@@ -372,7 +369,7 @@ static PyObject* diff(PyObject* self, PyObject* args)
         return NULL;
     }
     PyTuple_SET_ITEM(results, 0, controlTuples);
-    temp = PyString_FromStringAndSize((char *) db, dblen);
+    temp = PyBytes_FromStringAndSize((char *) db, dblen);
     PyMem_Free(db);
     if (!temp) {
         PyMem_Free(eb);
@@ -380,7 +377,7 @@ static PyObject* diff(PyObject* self, PyObject* args)
         return NULL;
     }
     PyTuple_SET_ITEM(results, 1, temp);
-    temp = PyString_FromStringAndSize((char *) eb, eblen);
+    temp = PyBytes_FromStringAndSize((char *) eb, eblen);
     PyMem_Free(eb);
     if (!temp) {
         Py_DECREF(results);
@@ -395,19 +392,20 @@ static PyObject* diff(PyObject* self, PyObject* args)
 /* takes the original data and the control, diff and extra blocks produced
    by bsdiff and returns the new data
 */
-static PyObject* patch(PyObject* self, PyObject* args)
+static PyObject *
+patch(PyObject* self, PyObject* args)
 {
     char *origData, *newData, *diffBlock, *extraBlock, *diffPtr, *extraPtr;
-    int origDataLength, newDataLength, diffBlockLength, extraBlockLength;
+    Py_ssize_t origDataLength, newDataLength, diffBlockLength, extraBlockLength;
     PyObject *controlTuples, *tuple, *results;
     off_t oldpos, newpos, x, y, z;
     int i, j, numTuples;
 
-    if (!PyArg_ParseTuple(args, "s#iO!s#s#",
-                          &origData, &origDataLength,
-                          &newDataLength, &PyList_Type, &controlTuples,
-                          &diffBlock, &diffBlockLength, &extraBlock,
-                          &extraBlockLength))
+    if (!PyArg_ParseTuple(args, "s#nO!s#s#",
+                          &origData, &origDataLength, &newDataLength,
+                          &PyList_Type, &controlTuples,
+                          &diffBlock, &diffBlockLength,
+                          &extraBlock, &extraBlockLength))
         return NULL;
 
     /* allocate the memory for the new data */
@@ -436,8 +434,7 @@ static PyObject* patch(PyObject* self, PyObject* args)
         y = PyLong_AsLong(PyTuple_GET_ITEM(tuple, 1));
         z = PyLong_AsLong(PyTuple_GET_ITEM(tuple, 2));
         if (newpos + x > newDataLength ||
-                diffPtr + x > diffBlock + diffBlockLength ||
-                extraPtr + y > extraBlock + extraBlockLength) {
+                diffPtr + x > diffBlock + diffBlockLength) {
             PyMem_Free(newData);
             PyErr_SetString(PyExc_ValueError, "corrupt patch (overflow)");
             return NULL;
@@ -449,6 +446,12 @@ static PyObject* patch(PyObject* self, PyObject* args)
                 newData[newpos + j] += origData[oldpos + j];
         newpos += x;
         oldpos += x;
+        if (newpos + y > newDataLength ||
+                extraPtr + y > extraBlock + extraBlockLength) {
+            PyMem_Free(newData);
+            PyErr_SetString(PyExc_ValueError, "corrupt patch (overflow)");
+            return NULL;
+        }
         memcpy(newData + newpos, extraPtr, y);
         extraPtr += y;
         newpos += y;
@@ -464,14 +467,15 @@ static PyObject* patch(PyObject* self, PyObject* args)
         return NULL;
     }
 
-    results = PyString_FromStringAndSize(newData, newDataLength);
+    results = PyBytes_FromStringAndSize(newData, newDataLength);
     PyMem_Free(newData);
     return results;
 }
 
 
-/* encode an integer value as a string of 8 bytes */
-static PyObject *encode_int64(PyObject *self, PyObject *value)
+/* encode an integer value as 8 bytes */
+static PyObject *
+encode_int64(PyObject *self, PyObject *value)
 {
     long long x;
     char bs[8], sign = 0x00;
@@ -489,26 +493,27 @@ static PyObject *encode_int64(PyObject *self, PyObject *value)
         x >>= 8;  /* x /= 256 */
     }
     bs[7] |= sign;
-    return PyString_FromStringAndSize(bs, 8);
+    return PyBytes_FromStringAndSize(bs, 8);
 }
 
 
-/* decode an off_t value from an 8 byte string */
-static PyObject *decode_int64(PyObject *self, PyObject *string)
+/* decode an off_t value from 8 bytes */
+static PyObject *
+decode_int64(PyObject *self, PyObject *string)
 {
     long long x;
     char *bs;
     int i;
 
-    if (!PyString_Check(string)) {
-        PyErr_SetString(PyExc_TypeError, "string expected");
+    if (!PyBytes_Check(string)) {
+        PyErr_SetString(PyExc_TypeError, "bytes expected");
         return NULL;
     }
-    if (PyString_Size(string) != 8) {
+    if (PyBytes_Size(string) != 8) {
         PyErr_SetString(PyExc_ValueError, "8 bytes expected");
         return NULL;
     }
-    bs = PyString_AsString(string);
+    bs = PyBytes_AsString(string);
 
     x = bs[7] & 0x7F;
     for (i = 6; i >= 0; i--) {
@@ -523,15 +528,15 @@ static PyObject *decode_int64(PyObject *self, PyObject *string)
 
 /* declaration of methods supported by this module */
 static PyMethodDef module_functions[] = {
-    {"diff", diff, METH_VARARGS},
-    {"patch", patch, METH_VARARGS},
+    {"diff",         diff,         METH_VARARGS},
+    {"patch",        patch,        METH_VARARGS},
     {"encode_int64", encode_int64, METH_O},
     {"decode_int64", decode_int64, METH_O},
     {NULL, NULL, 0, NULL}  /* Sentinel */
 };
 
 /* initialization routine for the shared libary */
-#ifdef IS_PY3K
+#if PY_MAJOR_VERSION == 3
 static PyModuleDef moduledef = {
     PyModuleDef_HEAD_INIT, "core", 0, -1, module_functions,
 };
